# define PARALLEL /* 並列探索を有効にする */
# define N_THREAD 2

# include <stdlib.h>
# include <stdio.h>
# include "parallelizer.h"

# ifdef PARALLEL
#  include <pthread.h>

# define CHECK(exp) check(#exp, exp)

static void check(const char *msg, int val)
{
  if(val)
  {
    fprintf(stderr, "%d: %s\n", val, msg);
    exit(1);
  }
}

struct tag_parallelizer
{
  pthread_t threads[N_THREAD];
  pthread_mutex_t work_mutex;
  pthread_cond_t work_available;
  pthread_cond_t work_consumed;
  void (*work_fun)(void *);
  void *work_param;
};

static void *thread_body(void *param)
{
  parallelizer *p = (parallelizer *)param;

  while(1)
  {
    void (*work_fun)(void *);
    void *work_param;
    CHECK(pthread_mutex_lock(&p->work_mutex));
    while(!p->work_fun)
      CHECK(pthread_cond_wait(&p->work_available, &p->work_mutex));
    work_fun = p->work_fun;
    work_param = p->work_param;
    p->work_fun = 0;
    CHECK(pthread_cond_signal(&p->work_consumed));
    CHECK(pthread_mutex_unlock(&p->work_mutex));
    work_fun(work_param);
  }
  return NULL;
}

parallelizer *new_pz(void)
{
  int i;
  parallelizer *p = malloc(sizeof(parallelizer));

  CHECK(pthread_mutex_init(&p->work_mutex, NULL));
  CHECK(pthread_cond_init(&p->work_available, NULL));
  CHECK(pthread_cond_init(&p->work_consumed, NULL));
  p->work_fun = 0;

  for(i = 0; i < N_THREAD; i++)
    CHECK(pthread_create(&p->threads[i], NULL, thread_body, p));
  return p;
}

void do_pz(parallelizer *p, void (*fun)(void *), void *param)
{
  CHECK(pthread_mutex_lock(&p->work_mutex));
  while(p->work_fun)
    CHECK(pthread_cond_wait(&p->work_consumed, &p->work_mutex));
  p->work_fun = fun;
  p->work_param = param;
  CHECK(pthread_cond_signal(&p->work_available));
  CHECK(pthread_mutex_unlock(&p->work_mutex));
}

void wait_and_destroy_pz(parallelizer *p)
{
  int i;
  for(i = 0; i < N_THREAD; i++)
    do_pz(p, pthread_exit, NULL);
  for(i = 0; i < N_THREAD; i++)
    CHECK(pthread_join(p->threads[i], NULL));
  pthread_mutex_destroy(&p->work_mutex);
  pthread_cond_destroy(&p->work_available);
  pthread_cond_destroy(&p->work_consumed);
  free(p);
}

# else

struct tag_parallelizer {};

parallelizer *new_pz(void) { return (parallelizer *)malloc(sizeof(parallelizer)); }
void do_pz(parallelizer *p, void (*fun)(void *), void *param) { fun(param); }
void wait_and_destroy_pz(parallelizer *p) { free(p); }

# endif
